@@ -274,7 +274,9 @@ static const SECTION union { const guint8 data[2069]; const double alignment; vo
 #endif /* !_MSC_VER */
 
 static GStaticResource static_resource = { na_resource_data.data, sizeof (na_resource_data.data) - 1 /* nul terminator */, NULL, NULL, NULL };
-extern GResource *na_get_resource (void);
+
+G_MODULE_EXPORT
+GResource *na_get_resource (void);
 GResource *na_get_resource (void)
 {
   return g_static_resource_get_resource (&static_resource);
@@ -323,13 +325,13 @@ GResource *na_get_resource (void)
  */
 
 /* We need to account for differences between the mangling of symbols
- * for Win32 (x86) and x64 programs, as symbols on Win32 are prefixed
- * with an underscore but symbols on x64 are not.
+ * for x86 and x64/ARM/ARM64 programs, as symbols on x86 are prefixed
+ * with an underscore but symbols on x64/ARM/ARM64 are not.
  */
-#ifdef _WIN64
-#define G_MSVC_SYMBOL_PREFIX ""
-#else
+#ifdef _M_IX86
 #define G_MSVC_SYMBOL_PREFIX "_"
+#else
+#define G_MSVC_SYMBOL_PREFIX ""
 #endif
 
 #define G_DEFINE_CONSTRUCTOR(_func) G_MSVC_CTOR (_func, G_MSVC_SYMBOL_PREFIX)
