@@ -8,9 +8,7 @@
  *
  */
 
-#ifdef HAVE_CONFIG_H
-	#include <config.h>
-#endif
+#include <config.h>
 
 #include <string.h>
 
@@ -19,12 +17,18 @@
 
 #include <glib/gi18n.h>
 #include <gtk/gtk.h>
+#include <gio/gio.h>
+
+#ifdef HAVE_X11
+#include <gdk/gdkx.h>
 #define WNCK_I_KNOW_THIS_IS_UNSTABLE
 #include <libwnck/libwnck.h>
-#include <gio/gio.h>
-#ifdef HAVE_WINDOW_PREVIEWS
-#include <gdk/gdkx.h>
-#endif
+#endif /* HAVE_X11 */
+
+#ifdef HAVE_WAYLAND
+#include <gdk/gdkwayland.h>
+#include "wayland-backend.h"
+#endif /* HAVE_WAYLAND */
 
 #define MATE_DESKTOP_USE_UNSTABLE_API
 #include <libmate-desktop/mate-desktop-utils.h>
@@ -34,22 +38,23 @@
 
 #define WINDOW_LIST_ICON "mate-panel-window-list"
 #define WINDOW_LIST_SCHEMA "org.mate.panel.applet.window-list"
+
 #ifdef HAVE_WINDOW_PREVIEWS
 #define WINDOW_LIST_PREVIEW_SCHEMA "org.mate.panel.applet.window-list-previews"
-#endif
+#endif /* HAVE_WINDOW_PREVIEWS */
+
+enum TasklistGroupingType {
+  TASKLIST_NEVER_GROUP,
+  TASKLIST_AUTO_GROUP,
+  TASKLIST_ALWAYS_GROUP
+};
 
 typedef struct {
 	GtkWidget* applet;
 	GtkWidget* tasklist;
 #ifdef HAVE_WINDOW_PREVIEWS
 	GtkWidget* preview;
-
-	gboolean show_window_thumbnails;
-	gint thumbnail_size;
 #endif
-	gboolean include_all_workspaces;
-	WnckTasklistGroupingType grouping;
-	gboolean move_unminimized_windows;
 
 	GtkOrientation orientation;
 	int size;
@@ -61,19 +66,22 @@ typedef struct {
 
 	/* Properties: */
 	GtkWidget* properties_dialog;
-	GtkWidget* show_current_radio;
-	GtkWidget* show_all_radio;
+	GtkWidget* display_all_workspaces_radio;
 #ifdef HAVE_WINDOW_PREVIEWS
-	GtkWidget* show_thumbnails_radio;
-	GtkWidget* hide_thumbnails_radio;
+	GtkWidget* window_thumbnail_box;
+	GtkWidget* show_thumbnails_check;
+	GtkWidget* thumbnail_size_label;
 	GtkWidget* thumbnail_size_spin;
+	GtkWidget* thumbnail_box;
 #endif
 	GtkWidget* never_group_radio;
 	GtkWidget* auto_group_radio;
 	GtkWidget* always_group_radio;
-	GtkWidget* minimized_windows_label;
-	GtkWidget* move_minimized_radio;
-	GtkWidget* change_workspace_radio;
+	GtkWidget* restore_to_current_workspace_radio;
+	GtkWidget* mouse_scroll_check;
+	GtkWidget* minimized_windows_box;
+	GtkWidget* window_grouping_box;
+	GtkWidget* window_list_content_box;
 
 	GSettings* settings;
 #ifdef HAVE_WINDOW_PREVIEWS
@@ -87,7 +95,7 @@ static void display_help_dialog(GtkAction* action, TasklistData* tasklist);
 static void display_about_dialog(GtkAction* action, TasklistData* tasklist);
 static void destroy_tasklist(GtkWidget* widget, TasklistData* tasklist);
 
-static void tasklist_update(TasklistData* tasklist)
+static void tasklist_set_size_request(TasklistData* tasklist)
 {
 	if (tasklist->orientation == GTK_ORIENTATION_HORIZONTAL)
 	{
@@ -97,13 +105,55 @@ static void tasklist_update(TasklistData* tasklist)
 	{
 		gtk_widget_set_size_request(GTK_WIDGET(tasklist->tasklist), tasklist->size, -1);
 	}
-
-	wnck_tasklist_set_grouping(WNCK_TASKLIST(tasklist->tasklist), tasklist->grouping);
-	wnck_tasklist_set_include_all_workspaces(WNCK_TASKLIST(tasklist->tasklist), tasklist->include_all_workspaces);
-	wnck_tasklist_set_switch_workspace_on_unminimize(WNCK_TASKLIST(tasklist->tasklist), tasklist->move_unminimized_windows);
 }
 
-static void response_cb(GtkWidget* widget, int id, TasklistData* tasklist)
+static void tasklist_apply_orientation(TasklistData* tasklist)
+{
+#ifdef HAVE_X11
+	if (WNCK_IS_TASKLIST(tasklist->tasklist))
+	{
+		wnck_tasklist_set_orientation(WNCK_TASKLIST(tasklist->tasklist), tasklist->orientation);
+	}
+#endif /* HAVE_X11 */
+
+#ifdef HAVE_WAYLAND
+	if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default()))
+	{
+		wayland_tasklist_set_orientation(tasklist->tasklist, tasklist->orientation);
+	}
+#endif
+}
+
+static void tasklist_set_button_relief(TasklistData* tasklist, GtkReliefStyle relief)
+{
+#ifdef HAVE_X11
+	if (WNCK_IS_TASKLIST(tasklist->tasklist))
+	{
+		wnck_tasklist_set_button_relief(WNCK_TASKLIST(tasklist->tasklist), relief);
+	}
+#endif /* HAVE_X11 */
+
+	/* Not implemented for Wayland */
+}
+
+static const int* tasklist_get_size_hint_list(TasklistData* tasklist, int* n_elements)
+{
+#ifdef HAVE_X11
+	if (WNCK_IS_TASKLIST(tasklist->tasklist))
+	{
+		return wnck_tasklist_get_size_hint_list(WNCK_TASKLIST(tasklist->tasklist), n_elements);
+	}
+	else
+#endif /* HAVE_X11 */
+
+	{
+		/* Not implemented for Wayland */
+		*n_elements = 0;
+		return NULL;
+	}
+}
+
+static void on_tasklist_properties_dialog_response(GtkWidget* widget, int id, TasklistData* tasklist)
 {
 	if (id == GTK_RESPONSE_HELP)
 	{
@@ -141,9 +191,8 @@ static void applet_change_orient(MatePanelApplet* applet, MatePanelAppletOrient
 		return;
 
 	tasklist->orientation = new_orient;
-	wnck_tasklist_set_orientation (WNCK_TASKLIST (tasklist->tasklist), new_orient);
-
-	tasklist_update(tasklist);
+	tasklist_apply_orientation (tasklist);
+	tasklist_set_size_request(tasklist);
 }
 
 static void applet_change_background(MatePanelApplet* applet, MatePanelAppletBackgroundType type, GdkColor* color, cairo_pattern_t* pattern, TasklistData* tasklist)
@@ -153,77 +202,95 @@ static void applet_change_background(MatePanelApplet* applet, MatePanelAppletBac
 		case PANEL_NO_BACKGROUND:
 		case PANEL_COLOR_BACKGROUND:
 		case PANEL_PIXMAP_BACKGROUND:
-			wnck_tasklist_set_button_relief(WNCK_TASKLIST(tasklist->tasklist), GTK_RELIEF_NONE);
+			tasklist_set_button_relief(tasklist, GTK_RELIEF_NONE);
 			break;
 	}
 }
 
+#ifdef HAVE_X11
 #ifdef HAVE_WINDOW_PREVIEWS
-static GdkPixbuf *preview_window_thumbnail (WnckWindow *wnck_window, TasklistData *tasklist)
+static cairo_surface_t*
+preview_window_thumbnail (WnckWindow   *wnck_window,
+                          TasklistData *tasklist,
+                          int          *thumbnail_width,
+                          int          *thumbnail_height,
+                          int          *thumbnail_scale)
 {
 	GdkWindow *window;
-	GdkPixbuf *screenshot;
-	GdkPixbuf *thumbnail;
+	GdkWindow *window_wrapper = NULL;
+	Window win;
+	cairo_surface_t *thumbnail;
+	cairo_t *cr;
 	double ratio;
-	int width, height;
-	int scale;
+	int width, height, scale;
 
-	window = gdk_x11_window_foreign_new_for_display (gdk_display_get_default (), wnck_window_get_xid (wnck_window));
+	win = wnck_window_get_xid (wnck_window);
 
-	if (window == NULL)
-		return NULL;
-
-	scale = gdk_window_get_scale_factor (window);
-	width = gdk_window_get_width (window) * scale;
-	height = gdk_window_get_height (window) * scale;
-
-	/* Generate window screenshot for preview */
-	screenshot = gdk_pixbuf_get_from_window (window, 0, 0, width / scale, height / scale);
-	g_object_unref (window);
-
-	if (screenshot == NULL)
-		return NULL;
-
-	/* Determine whether the contents of the screenshot are empty */
-	if (gdk_pixbuf_get_byte_length (screenshot) == 0)
+	if ((window = gdk_x11_window_lookup_for_display (gdk_display_get_default (), win)) == NULL)
 	{
-		g_object_unref (screenshot);
-		return NULL;
-	}
-
-	/* Scale to configured size while maintaining aspect ratio */
-	if (width > height)
-	{
-		ratio = (double) height / (double) width;
-		width = MIN(width, tasklist->thumbnail_size);
-		height = width * ratio;
+		if ((window = gdk_x11_window_foreign_new_for_display (gdk_display_get_default (), win)) == NULL)
+		{
+			return NULL;
+		}
+		else
+		{
+			window_wrapper = window;
+		}
 	}
 	else
 	{
-		ratio = (double) width / (double) height;
-		height = MIN(height, tasklist->thumbnail_size);
-		width = height * ratio;
+		g_object_ref (window);
 	}
 
-	thumbnail = gdk_pixbuf_scale_simple (screenshot, width, height, GDK_INTERP_BILINEAR);
-	g_object_unref (screenshot);
+	*thumbnail_scale = scale = gdk_window_get_scale_factor (window);
+	width = gdk_window_get_width (window) * scale;
+	height = gdk_window_get_height (window) * scale;
+
+	/* Scale to configured size while maintaining aspect ratio */
+	gint thumbnail_size = g_settings_get_int (tasklist->preview_settings, "thumbnail-window-size");
+	if (width > height)
+	{
+		int max_size = MIN (width, thumbnail_size * scale);
+		ratio = (double) max_size / (double) width;
+		*thumbnail_width = max_size;
+		*thumbnail_height = (int) ((double) height * ratio);
+	}
+	else
+	{
+		int max_size = MIN (height, thumbnail_size * scale);
+		ratio = (double) max_size / (double) height;
+		*thumbnail_height = max_size;
+		*thumbnail_width = (int) ((double) width * ratio);
+	}
+
+	thumbnail = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
+	                                        *thumbnail_width,
+	                                        *thumbnail_height);
+	cairo_surface_set_device_scale (thumbnail, scale, scale);
+	cr = cairo_create (thumbnail);
+	cairo_scale (cr, ratio, ratio);
+	gdk_cairo_set_source_window (cr, window, 0, 0);
+	cairo_paint (cr);
+	cairo_destroy (cr);
+
+	if (window_wrapper)
+		g_object_unref (window_wrapper);
+	g_object_unref (window);
 
 	return thumbnail;
 }
 
 #define PREVIEW_PADDING 5
-static void preview_window_reposition (TasklistData *tasklist, GdkPixbuf *thumbnail)
+static void
+preview_window_reposition (TasklistData    *tasklist,
+                           cairo_surface_t *thumbnail,
+                           int              width,
+                           int              height,
+                           int              scale)
 {
 	GdkMonitor *monitor;
 	GdkRectangle monitor_geom;
 	int x_pos, y_pos;
-	int width, height;
-
-	width = gdk_pixbuf_get_width (thumbnail);
-	height = gdk_pixbuf_get_height (thumbnail);
-
-	/* Resize window to fit thumbnail */
-	gtk_window_resize (GTK_WINDOW (tasklist->preview), width, height);
 
 	/* Set position at pointer, then re-adjust from there to just outside of the pointer */
 	gtk_window_set_position (GTK_WINDOW (tasklist->preview), GTK_WIN_POS_MOUSE);
@@ -237,13 +304,13 @@ static void preview_window_reposition (TasklistData *tasklist, GdkPixbuf *thumbn
 	switch (mate_panel_applet_get_orient (MATE_PANEL_APPLET (tasklist->applet)))
 	{
 		case MATE_PANEL_APPLET_ORIENT_LEFT:
-			x_pos = monitor_geom.width + monitor_geom.x - (width + tasklist->size) - PREVIEW_PADDING;
+			x_pos = monitor_geom.width + monitor_geom.x - (width/scale + tasklist->size) - PREVIEW_PADDING;
 			break;
 		case MATE_PANEL_APPLET_ORIENT_RIGHT:
 			x_pos = tasklist->size + PREVIEW_PADDING;
 			break;
 		case MATE_PANEL_APPLET_ORIENT_UP:
-			y_pos = monitor_geom.height + monitor_geom.y - (height + tasklist->size) - PREVIEW_PADDING;
+			y_pos = monitor_geom.height + monitor_geom.y - (height/scale + tasklist->size) - PREVIEW_PADDING;
 			break;
 		case MATE_PANEL_APPLET_ORIENT_DOWN:
 		default:
@@ -254,21 +321,24 @@ static void preview_window_reposition (TasklistData *tasklist, GdkPixbuf *thumbn
 	gtk_window_move (GTK_WINDOW (tasklist->preview), x_pos, y_pos);
 }
 
-static gboolean preview_window_draw (GtkWidget *widget, cairo_t *cr, GdkPixbuf *thumbnail)
+static gboolean preview_window_draw (GtkWidget *widget, cairo_t *cr, cairo_surface_t *thumbnail)
 {
 	GtkStyleContext *context;
 
 	context = gtk_widget_get_style_context (widget);
-	gtk_render_icon (context, cr, thumbnail, 0, 0);
+	gtk_render_icon_surface (context, cr, thumbnail, 0, 0);
 
 	return FALSE;
 }
 
 static gboolean applet_enter_notify_event (WnckTasklist *tl, GList *wnck_windows, TasklistData *tasklist)
 {
-	GdkPixbuf *thumbnail;
+	cairo_surface_t *thumbnail;
 	WnckWindow *wnck_window = NULL;
 	int n_windows;
+	int thumbnail_width;
+	int thumbnail_height;
+	int thumbnail_scale;
 
 	if (tasklist->preview != NULL)
 	{
@@ -276,7 +346,7 @@ static gboolean applet_enter_notify_event (WnckTasklist *tl, GList *wnck_windows
 		tasklist->preview = NULL;
 	}
 
-	if (!tasklist->show_window_thumbnails || wnck_windows == NULL)
+	if (!g_settings_get_boolean (tasklist->preview_settings, "show-window-thumbnails") || wnck_windows == NULL)
 		return FALSE;
 
 	n_windows = g_list_length (wnck_windows);
@@ -296,7 +366,7 @@ static gboolean applet_enter_notify_event (WnckTasklist *tl, GList *wnck_windows
 						  wnck_screen_get_active_workspace (wnck_screen_get_default ())))
 		return FALSE;
 
-	thumbnail = preview_window_thumbnail (wnck_window, tasklist);
+	thumbnail = preview_window_thumbnail (wnck_window, tasklist, &thumbnail_width, &thumbnail_height, &thumbnail_scale);
 
 	if (thumbnail == NULL)
 		return FALSE;
@@ -305,13 +375,14 @@ static gboolean applet_enter_notify_event (WnckTasklist *tl, GList *wnck_windows
 	tasklist->preview = gtk_window_new (GTK_WINDOW_POPUP);
 
 	gtk_widget_set_app_paintable (tasklist->preview, TRUE);
+	gtk_window_set_default_size (GTK_WINDOW (tasklist->preview), thumbnail_width/thumbnail_scale, thumbnail_height/thumbnail_scale);
 	gtk_window_set_resizable (GTK_WINDOW (tasklist->preview), TRUE);
 
-	preview_window_reposition (tasklist, thumbnail);
+	preview_window_reposition (tasklist, thumbnail, thumbnail_width, thumbnail_height, thumbnail_scale);
 
 	gtk_widget_show (tasklist->preview);
 
-	g_signal_connect_data (G_OBJECT (tasklist->preview), "draw", G_CALLBACK (preview_window_draw), thumbnail, (GClosureNotify) g_object_unref, 0);
+	g_signal_connect_data (G_OBJECT (tasklist->preview), "draw", G_CALLBACK (preview_window_draw), thumbnail, (GClosureNotify) G_CALLBACK (cairo_surface_destroy), 0);
 
 	return FALSE;
 }
@@ -326,7 +397,8 @@ static gboolean applet_leave_notify_event (WnckTasklist *tl, GList *wnck_windows
 
 	return FALSE;
 }
-#endif
+#endif /* HAVE_WINDOW_PREVIEWS */
+#endif /* HAVE_X11 */
 
 static void applet_change_pixel_size(MatePanelApplet* applet, gint size, TasklistData* tasklist)
 {
@@ -335,7 +407,7 @@ static void applet_change_pixel_size(MatePanelApplet* applet, gint size, Tasklis
 
 	tasklist->size = size;
 
-	tasklist_update(tasklist);
+	tasklist_set_size_request(tasklist);
 }
 
 /* TODO: this is sad, should be used a function to retrieve  applications from
@@ -380,198 +452,12 @@ static const GtkActionEntry tasklist_menu_actions[] = {
 	}
 };
 
-static void tasklist_properties_update_content_radio(TasklistData* tasklist)
-{
-	GtkWidget* button;
-
-	if (tasklist->show_current_radio == NULL)
-		return;
-
-	if (tasklist->include_all_workspaces)
-	{
-		button = tasklist->show_all_radio;
-	}
-	else
-	{
-		button = tasklist->show_current_radio;
-	}
-
-	if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)))
-		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
-
-	gtk_widget_set_sensitive(tasklist->minimized_windows_label, tasklist->include_all_workspaces);
-	gtk_widget_set_sensitive(tasklist->move_minimized_radio, tasklist->include_all_workspaces);
-	gtk_widget_set_sensitive(tasklist->change_workspace_radio, tasklist->include_all_workspaces);
-}
-
-static void display_all_workspaces_changed(GSettings* settings, gchar* key, TasklistData* tasklist)
-{
-	gboolean value;
-
-	value = g_settings_get_boolean(settings, key);
-
-	tasklist->include_all_workspaces = (value != 0);
-	tasklist_update(tasklist);
-
-	tasklist_properties_update_content_radio(tasklist);
-}
-
-#ifdef HAVE_WINDOW_PREVIEWS
-static void tasklist_update_thumbnails_radio(TasklistData* tasklist)
-{
-	GtkWidget* button;
-
-	if (tasklist->show_thumbnails_radio == NULL || tasklist->hide_thumbnails_radio == NULL)
-		return;
-
-	if (tasklist->show_window_thumbnails)
-	{
-		button = tasklist->show_thumbnails_radio;
-	}
-	else
-	{
-		button = tasklist->hide_thumbnails_radio;
-	}
-
-	if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)))
-		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
-}
-
-static void window_thumbnails_changed(GSettings *settings, gchar* key, TasklistData* tasklist)
-{
-	gboolean value;
-
-	value = g_settings_get_boolean(settings, key);
-
-	tasklist->show_window_thumbnails = (value != 0);
-
-	tasklist_update_thumbnails_radio(tasklist);
-}
-
-static void tasklist_update_thumbnail_size_spin(TasklistData* tasklist)
-{
-	GtkWidget* button;
-
-	if (!tasklist->thumbnail_size)
-		return;
-
-	button = tasklist->thumbnail_size_spin;
-
-	gtk_spin_button_set_value(GTK_SPIN_BUTTON(button), (gdouble)tasklist->thumbnail_size);
-}
-
-static void thumbnail_size_changed(GSettings *settings, gchar* key, TasklistData* tasklist)
-{
-	tasklist->thumbnail_size = g_settings_get_int(settings, key);
-	tasklist_update_thumbnail_size_spin(tasklist);
-}
-#endif
-
-static GtkWidget* get_grouping_button(TasklistData* tasklist, WnckTasklistGroupingType type)
-{
-	switch (type)
-	{
-		default:
-		case WNCK_TASKLIST_NEVER_GROUP:
-			return tasklist->never_group_radio;
-			break;
-		case WNCK_TASKLIST_AUTO_GROUP:
-			return tasklist->auto_group_radio;
-			break;
-		case WNCK_TASKLIST_ALWAYS_GROUP:
-			return tasklist->always_group_radio;
-			break;
-	}
-}
-
-static void group_windows_changed(GSettings* settings, gchar* key, TasklistData* tasklist)
-{
-	WnckTasklistGroupingType type;
-	GtkWidget* button;
-
-	type = g_settings_get_enum (settings, key);
-
-	tasklist->grouping = type;
-	tasklist_update(tasklist);
-
-	button = get_grouping_button(tasklist, type);
-
-	if (button && !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)))
-	{
-		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
-	}
-}
-
-static void tasklist_update_unminimization_radio(TasklistData* tasklist)
-{
-	GtkWidget* button;
-
-	if (tasklist->move_minimized_radio == NULL)
-		return;
-
-	if (tasklist->move_unminimized_windows)
-	{
-		button = tasklist->move_minimized_radio;
-	}
-	else
-	{
-		button = tasklist->change_workspace_radio;
-	}
-
-	if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)))
-		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
-}
-
-
-static void move_unminimized_windows_changed(GSettings* settings, gchar* key, TasklistData* tasklist)
-{
-	gboolean value;
-
-	value = g_settings_get_boolean(settings, key);
-
-	tasklist->move_unminimized_windows = (value != 0);
-	tasklist_update(tasklist);
-
-	tasklist_update_unminimization_radio(tasklist);
-}
-
-static void setup_gsettings(TasklistData* tasklist)
-{
-	tasklist->settings = mate_panel_applet_settings_new (MATE_PANEL_APPLET (tasklist->applet), WINDOW_LIST_SCHEMA);
-
-	g_signal_connect (tasklist->settings,
-					  "changed::display-all-workspaces",
-					  G_CALLBACK (display_all_workspaces_changed),
-					  tasklist);
-
-#ifdef HAVE_WINDOW_PREVIEWS
-	tasklist->preview_settings = mate_panel_applet_settings_new (MATE_PANEL_APPLET (tasklist->applet), WINDOW_LIST_PREVIEW_SCHEMA);
-
-	g_signal_connect (tasklist->preview_settings,
-					  "changed::show-window-thumbnails",
-					  G_CALLBACK (window_thumbnails_changed),
-					  tasklist);
-	g_signal_connect (tasklist->preview_settings,
-					  "changed::thumbnail-window-size",
-					  G_CALLBACK (thumbnail_size_changed),
-					  tasklist);
-#endif
-	g_signal_connect (tasklist->settings,
-					  "changed::group-windows",
-					  G_CALLBACK (group_windows_changed),
-					  tasklist);
-	g_signal_connect (tasklist->settings,
-					  "changed::move-unminimized-windows",
-					  G_CALLBACK (move_unminimized_windows_changed),
-					  tasklist);
-}
-
 static void applet_size_allocate(GtkWidget *widget, GtkAllocation *allocation, TasklistData *tasklist)
 {
 	int len;
 	const int* size_hints;
 
-	size_hints = wnck_tasklist_get_size_hint_list (WNCK_TASKLIST (tasklist->tasklist), &len);
+	size_hints = tasklist_get_size_hint_list (tasklist, &len);
 
 	g_assert(len % 2 == 0);
 
@@ -598,6 +484,8 @@ static void applet_size_allocate(GtkWidget *widget, GtkAllocation *allocation, T
 		mate_panel_applet_set_size_hints(MATE_PANEL_APPLET(tasklist->applet), size_hints, len, 0);
 }
 
+#ifdef HAVE_X11
+/* Currently only used on X11, but should work on Wayland as well when needed */
 static GdkPixbuf* icon_loader_func(const char* icon, int size, unsigned int flags, void* data)
 {
 	TasklistData* tasklist;
@@ -643,6 +531,7 @@ static GdkPixbuf* icon_loader_func(const char* icon, int size, unsigned int flag
 
 	return retval;
 }
+#endif /* HAVE_X11 */
 
 gboolean window_list_applet_fill(MatePanelApplet* applet)
 {
@@ -670,20 +559,12 @@ gboolean window_list_applet_fill(MatePanelApplet* applet)
 
 	mate_panel_applet_set_flags(MATE_PANEL_APPLET(tasklist->applet), MATE_PANEL_APPLET_EXPAND_MAJOR | MATE_PANEL_APPLET_EXPAND_MINOR | MATE_PANEL_APPLET_HAS_HANDLE);
 
-	setup_gsettings(tasklist);
-
-	tasklist->include_all_workspaces = g_settings_get_boolean (tasklist->settings, "display-all-workspaces");
+	tasklist->settings = mate_panel_applet_settings_new (MATE_PANEL_APPLET (tasklist->applet), WINDOW_LIST_SCHEMA);
 
 #ifdef HAVE_WINDOW_PREVIEWS
-	tasklist->show_window_thumbnails = g_settings_get_boolean (tasklist->preview_settings, "show-window-thumbnails");
-
-	tasklist->thumbnail_size = g_settings_get_int (tasklist->preview_settings, "thumbnail-window-size");
+ 	tasklist->preview_settings = mate_panel_applet_settings_new (MATE_PANEL_APPLET (tasklist->applet), WINDOW_LIST_PREVIEW_SCHEMA);
 #endif
 
-	tasklist->grouping = g_settings_get_enum (tasklist->settings, "group-windows");
-
-	tasklist->move_unminimized_windows = g_settings_get_boolean (tasklist->settings, "move-unminimized-windows");
-
 	tasklist->size = mate_panel_applet_get_size(applet);
 
 #if !defined(WNCKLET_INPROCESS) && !GTK_CHECK_VERSION (3, 23, 0)
@@ -703,18 +584,37 @@ gboolean window_list_applet_fill(MatePanelApplet* applet)
 			break;
 	}
 
-	tasklist->tasklist = wnck_tasklist_new();
+#ifdef HAVE_X11
+	if (GDK_IS_X11_DISPLAY (gdk_display_get_default ()))
+	{
+		tasklist->tasklist = wnck_tasklist_new();
 
-	wnck_tasklist_set_orientation (WNCK_TASKLIST (tasklist->tasklist), tasklist->orientation);
-	wnck_tasklist_set_middle_click_close (WNCK_TASKLIST (tasklist->tasklist), TRUE);
-	wnck_tasklist_set_icon_loader(WNCK_TASKLIST(tasklist->tasklist), icon_loader_func, tasklist, NULL);
+		wnck_tasklist_set_middle_click_close (WNCK_TASKLIST (tasklist->tasklist), TRUE);
+		wnck_tasklist_set_icon_loader(WNCK_TASKLIST(tasklist->tasklist), icon_loader_func, tasklist, NULL);
+
+#ifdef HAVE_WINDOW_PREVIEWS
+		g_signal_connect(G_OBJECT(tasklist->tasklist), "task_enter_notify", G_CALLBACK(applet_enter_notify_event), tasklist);
+		g_signal_connect(G_OBJECT(tasklist->tasklist), "task_leave_notify", G_CALLBACK(applet_leave_notify_event), tasklist);
+#endif /* HAVE_WINDOW_PREVIEWS */
+	}
+	else
+#endif /* HAVE_X11 */
+
+#ifdef HAVE_WAYLAND
+	if (GDK_IS_WAYLAND_DISPLAY (gdk_display_get_default ()))
+	{
+		tasklist->tasklist = wayland_tasklist_new();
+	}
+	else
+#endif /* HAVE_WAYLAND */
+
+	{
+		tasklist->tasklist = gtk_label_new ("[Tasklist not supported on this platform]");
+	}
+
+	tasklist_apply_orientation(tasklist);
 
 	g_signal_connect(G_OBJECT(tasklist->tasklist), "destroy", G_CALLBACK(destroy_tasklist), tasklist);
-#ifdef HAVE_WINDOW_PREVIEWS
-	g_signal_connect(G_OBJECT(tasklist->tasklist), "task_enter_notify", G_CALLBACK(applet_enter_notify_event), tasklist);
-	g_signal_connect(G_OBJECT(tasklist->tasklist), "task_leave_notify", G_CALLBACK(applet_leave_notify_event), tasklist);
-#endif
-
 	g_signal_connect(G_OBJECT(tasklist->applet), "size_allocate", G_CALLBACK(applet_size_allocate), tasklist);
 
 	gtk_container_add(GTK_CONTAINER(tasklist->applet), tasklist->tasklist);
@@ -724,8 +624,6 @@ gboolean window_list_applet_fill(MatePanelApplet* applet)
 	g_signal_connect(G_OBJECT(tasklist->applet), "change_size", G_CALLBACK(applet_change_pixel_size), tasklist);
 	g_signal_connect(G_OBJECT(tasklist->applet), "change_background", G_CALLBACK(applet_change_background), tasklist);
 
-	mate_panel_applet_set_background_widget(MATE_PANEL_APPLET(tasklist->applet), GTK_WIDGET(tasklist->applet));
-
 	action_group = gtk_action_group_new("Tasklist Applet Actions");
 	gtk_action_group_set_translation_domain(action_group, GETTEXT_PACKAGE);
 	gtk_action_group_add_actions(action_group, tasklist_menu_actions, G_N_ELEMENTS(tasklist_menu_actions), tasklist);
@@ -771,7 +669,7 @@ gboolean window_list_applet_fill(MatePanelApplet* applet)
 
 	g_object_unref(action_group);
 
-	tasklist_update(tasklist);
+	tasklist_set_size_request(tasklist);
 	gtk_widget_show(tasklist->tasklist);
 	gtk_widget_show(tasklist->applet);
 
@@ -780,12 +678,11 @@ gboolean window_list_applet_fill(MatePanelApplet* applet)
 
 static void call_system_monitor(GtkAction* action, TasklistData* tasklist)
 {
-	char *programpath;
 	int i;
 
 	for (i = 0; i < G_N_ELEMENTS(system_monitors); i += 1)
 	{
-		programpath = g_find_program_in_path(system_monitors[i]);
+		char *programpath = g_find_program_in_path(system_monitors[i]);
 
 		if (programpath != NULL)
 		{
@@ -799,7 +696,6 @@ static void call_system_monitor(GtkAction* action, TasklistData* tasklist)
 	}
 }
 
-
 static void display_help_dialog(GtkAction* action, TasklistData* tasklist)
 {
 	wncklet_display_help(tasklist->applet, "mate-user-guide", "windowlist", WINDOW_LIST_ICON);
@@ -827,147 +723,128 @@ static void display_about_dialog(GtkAction* action, TasklistData* tasklist)
 		"comments", _("The Window List shows a list of all windows in a set of buttons and lets you browse them."),
 		"copyright", _("Copyright \xc2\xa9 2002 Red Hat, Inc.\n"
 		               "Copyright \xc2\xa9 2011 Perberos\n"
-		               "Copyright \xc2\xa9 2012-2020 MATE developers"),
+		               "Copyright \xc2\xa9 2012-2021 MATE developers"),
 		"documenters", documenters,
 		"icon-name", WINDOW_LIST_ICON,
 		"logo-icon-name", WINDOW_LIST_ICON,
 		"translator-credits", _("translator-credits"),
 		"version", VERSION,
-		"website", "http://www.mate-desktop.org/",
+		"website", PACKAGE_URL,
 		NULL);
 }
 
-static void group_windows_toggled(GtkToggleButton* button, TasklistData* tasklist)
+static void mouse_scrolling_callback(GSettings* settings, gchar *key, TasklistData* tasklist)
 {
-	if (gtk_toggle_button_get_active(button))
-	{
-		gchar *value;
-		value = g_object_get_data (G_OBJECT (button), "group_value");
-		g_settings_set_string (tasklist->settings, "group-windows", value);
-	}
+	wnck_tasklist_set_scroll_enabled (WNCK_TASKLIST(tasklist->tasklist), g_settings_get_boolean (settings, "scroll-enabled"));
+}
+
+static void move_unminimized_windows_callback(GSettings* settings, gchar *key, TasklistData* tasklist)
+{
+	wnck_tasklist_set_switch_workspace_on_unminimize(WNCK_TASKLIST(tasklist->tasklist), g_settings_get_boolean (settings, "move-unminimized-windows"));
+}
+
+static void display_all_workspaces_callback(GSettings* settings, gchar *key, TasklistData* tasklist)
+{
+	wnck_tasklist_set_include_all_workspaces(WNCK_TASKLIST(tasklist->tasklist), g_settings_get_boolean (settings, "display-all-workspaces"));
 }
 
 #ifdef HAVE_WINDOW_PREVIEWS
-static void show_thumbnails_toggled(GtkToggleButton* button, TasklistData* tasklist)
-{
-	g_settings_set_boolean(tasklist->preview_settings, "show-window-thumbnails", gtk_toggle_button_get_active(button));
-}
-
-static void thumbnail_size_spin_changed(GtkSpinButton* button, TasklistData* tasklist)
+static void on_thumbnail_size_spin_value_changed(GtkSpinButton* button, TasklistData* tasklist)
 {
 	g_settings_set_int(tasklist->preview_settings, "thumbnail-window-size", gtk_spin_button_get_value_as_int(button));
 }
 #endif
 
-static void move_minimized_toggled(GtkToggleButton* button, TasklistData* tasklist)
+static void on_never_group_radio_toggled(GtkRadioButton* button, TasklistData* tasklist)
 {
-	g_settings_set_boolean(tasklist->settings, "move-unminimized-windows", gtk_toggle_button_get_active(button));
+	g_settings_set_enum (tasklist->settings, "group-windows", TASKLIST_NEVER_GROUP);
+	wnck_tasklist_set_grouping(WNCK_TASKLIST(tasklist->tasklist), WNCK_TASKLIST_NEVER_GROUP);
 }
 
-static void display_all_workspaces_toggled(GtkToggleButton* button, TasklistData* tasklist)
+static void on_auto_group_radio_toggled(GtkRadioButton* button, TasklistData* tasklist)
 {
-	g_settings_set_boolean(tasklist->settings, "display-all-workspaces", gtk_toggle_button_get_active(button));
+	g_settings_set_enum (tasklist->settings, "group-windows", TASKLIST_AUTO_GROUP);
+	wnck_tasklist_set_grouping(WNCK_TASKLIST(tasklist->tasklist), WNCK_TASKLIST_AUTO_GROUP);
 }
 
-#define WID(s) GTK_WIDGET(gtk_builder_get_object(builder, s))
-
-static void setup_sensitivity(TasklistData* tasklist, GtkBuilder* builder, const char* wid1, const char* wid2, const char* wid3, const char* key)
+static void on_always_group_radio_toggled(GtkRadioButton* button, TasklistData* tasklist)
 {
-	GtkWidget* w;
-
-	if (g_settings_is_writable(tasklist->settings, key))
-	{
-		return;
-	}
-
-	w = WID(wid1);
-	g_assert(w != NULL);
-	gtk_widget_set_sensitive(w, FALSE);
-
-	if (wid2 != NULL)
-	{
-		w = WID(wid2);
-		g_assert(w != NULL);
-		gtk_widget_set_sensitive(w, FALSE);
-	}
-
-	if (wid3 != NULL)
-	{
-		w = WID(wid3);
-		g_assert(w != NULL);
-		gtk_widget_set_sensitive(w, FALSE);
-	}
+	g_settings_set_enum (tasklist->settings, "group-windows", TASKLIST_ALWAYS_GROUP);
+	wnck_tasklist_set_grouping(WNCK_TASKLIST(tasklist->tasklist), WNCK_TASKLIST_ALWAYS_GROUP);
 }
 
+#define GET_WIDGET(s) GTK_WIDGET(gtk_builder_get_object(builder, s))
+
 static void setup_dialog(GtkBuilder* builder, TasklistData* tasklist)
 {
-	GtkWidget* button;
-#ifdef HAVE_WINDOW_PREVIEWS
-	GtkAdjustment *adjustment;
-#endif
-
-	tasklist->show_current_radio = WID("show_current_radio");
-	tasklist->show_all_radio = WID("show_all_radio");
-
-	setup_sensitivity(tasklist, builder, "show_current_radio", "show_all_radio", NULL, "display-all-workspaces" /* key */);
-
-	tasklist->never_group_radio = WID("never_group_radio");
-	tasklist->auto_group_radio = WID("auto_group_radio");
-	tasklist->always_group_radio = WID("always_group_radio");
-
-	setup_sensitivity(tasklist, builder, "never_group_radio", "auto_group_radio", "always_group_radio", "group-windows" /* key */);
+	tasklist->display_all_workspaces_radio = GET_WIDGET("display_all_workspaces_radio");
+	tasklist->never_group_radio = GET_WIDGET("never_group_radio");
+	tasklist->auto_group_radio = GET_WIDGET("auto_group_radio");
+	tasklist->always_group_radio = GET_WIDGET("always_group_radio");
 
 #ifdef HAVE_WINDOW_PREVIEWS
-	tasklist->show_thumbnails_radio = WID("show_thumbnails_radio");
-	tasklist->hide_thumbnails_radio = WID("hide_thumbnails_radio");
-	tasklist->thumbnail_size_spin = WID("thumbnail_size_spin");
+	tasklist->window_thumbnail_box = GET_WIDGET("window_thumbnail_box");
+	tasklist->show_thumbnails_check = GET_WIDGET("show_thumbnails_check");
+	tasklist->thumbnail_size_label = GET_WIDGET("thumbnail_size_label");
+	tasklist->thumbnail_size_spin = GET_WIDGET("thumbnail_size_spin");
+	tasklist->thumbnail_box = GET_WIDGET("thumbnail_box");
 
-	setup_sensitivity(tasklist, builder, "show_thumbnails_radio", "hide_thumbnails_radio", NULL, "show-window-thumbnails" /* key */);
-	gtk_widget_set_sensitive(tasklist->thumbnail_size_spin, TRUE);
-	adjustment = gtk_spin_button_get_adjustment (GTK_SPIN_BUTTON(tasklist->thumbnail_size_spin));
-	gtk_adjustment_set_lower (adjustment, 0);
-	gtk_adjustment_set_upper (adjustment, 999);
-	gtk_adjustment_set_step_increment (adjustment, 1);
+	g_settings_bind(tasklist->preview_settings, "show-window-thumbnails",
+					tasklist->show_thumbnails_check, "active",
+					G_SETTINGS_BIND_DEFAULT);
+
+	gtk_widget_set_sensitive (tasklist->thumbnail_box, gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tasklist->show_thumbnails_check)));
+	g_object_bind_property(tasklist->show_thumbnails_check, "active", tasklist->thumbnail_box, "sensitive", G_BINDING_DEFAULT);
+
+	gtk_spin_button_set_value(GTK_SPIN_BUTTON(tasklist->thumbnail_size_spin),
+							  (gdouble) g_settings_get_int (tasklist->preview_settings, "thumbnail-window-size"));
 #else
-	gtk_widget_hide(WID("window_thumbnails"));
+	gtk_widget_hide(GET_WIDGET("window_thumbnail_box"));
 #endif
 
-	tasklist->minimized_windows_label = WID("minimized_windows_label");
-	tasklist->move_minimized_radio = WID("move_minimized_radio");
-	tasklist->change_workspace_radio = WID("change_workspace_radio");
+	tasklist->restore_to_current_workspace_radio = GET_WIDGET("restore_to_current_workspace_radio");
+	tasklist->mouse_scroll_check = GET_WIDGET("mouse_scroll_check");
+	tasklist->minimized_windows_box = GET_WIDGET("minimized_windows_box");
+	tasklist->window_grouping_box = GET_WIDGET("window_grouping_box");
+	tasklist->window_list_content_box = GET_WIDGET("window_list_content_box");
 
-	setup_sensitivity(tasklist, builder, "move_minimized_radio", "change_workspace_radio", NULL, "move-unminimized-windows" /* key */);
+	g_settings_bind (tasklist->settings, "scroll-enabled",
+					 tasklist->mouse_scroll_check, "active",
+					 G_SETTINGS_BIND_DEFAULT);
 
-	/* Window grouping: */
-	button = get_grouping_button(tasklist, tasklist->grouping);
-	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
-	g_object_set_data(G_OBJECT(tasklist->never_group_radio), "group_value", "never");
-	g_object_set_data(G_OBJECT(tasklist->auto_group_radio), "group_value", "auto");
-	g_object_set_data(G_OBJECT(tasklist->always_group_radio), "group_value", "always");
+	g_signal_connect (tasklist->settings,
+					  "changed::scroll-enabled",
+					  G_CALLBACK (mouse_scrolling_callback),
+					  tasklist);
 
-	g_signal_connect(G_OBJECT(tasklist->never_group_radio), "toggled", (GCallback) group_windows_toggled, tasklist);
-	g_signal_connect(G_OBJECT(tasklist->auto_group_radio), "toggled", (GCallback) group_windows_toggled, tasklist);
-	g_signal_connect(G_OBJECT(tasklist->always_group_radio), "toggled", (GCallback) group_windows_toggled, tasklist);
+	g_settings_bind (tasklist->settings, "move-unminimized-windows",
+					 tasklist->restore_to_current_workspace_radio, "active",
+					 G_SETTINGS_BIND_DEFAULT);
+	g_signal_connect (tasklist->settings, "changed::move-unminimized-windows",
+					  G_CALLBACK (move_unminimized_windows_callback),
+					  tasklist);
 
-#ifdef HAVE_WINDOW_PREVIEWS
-	/* show thumbnails on hover: */
-	tasklist_update_thumbnails_radio(tasklist);
-	g_signal_connect(G_OBJECT(tasklist->show_thumbnails_radio), "toggled", (GCallback) show_thumbnails_toggled, tasklist);
-	/* change thumbnail size: */
-	tasklist_update_thumbnail_size_spin(tasklist);
-	g_signal_connect(G_OBJECT(tasklist->thumbnail_size_spin), "value-changed", (GCallback) thumbnail_size_spin_changed, tasklist);
-#endif
+	g_settings_bind(tasklist->settings, "display-all-workspaces",
+					tasklist->display_all_workspaces_radio, "active",
+					G_SETTINGS_BIND_DEFAULT);
+	g_signal_connect (tasklist->settings, "changed::display-all-workspaces",
+					  G_CALLBACK (display_all_workspaces_callback),
+					  tasklist);
 
-	/* move window when unminimizing: */
-	tasklist_update_unminimization_radio(tasklist);
-	g_signal_connect(G_OBJECT(tasklist->move_minimized_radio), "toggled", (GCallback) move_minimized_toggled, tasklist);
+	gtk_widget_set_sensitive (tasklist->minimized_windows_box, gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tasklist->display_all_workspaces_radio)));
+	g_object_bind_property(tasklist->display_all_workspaces_radio, "active", tasklist->minimized_windows_box, "sensitive", G_BINDING_DEFAULT);
 
-	/* Tasklist content: */
-	tasklist_properties_update_content_radio (tasklist);
-	g_signal_connect(G_OBJECT(tasklist->show_all_radio), "toggled", (GCallback) display_all_workspaces_toggled, tasklist);
-
-	g_signal_connect_swapped(WID("done_button"), "clicked", (GCallback) gtk_widget_hide, tasklist->properties_dialog);
-	g_signal_connect(tasklist->properties_dialog, "response", G_CALLBACK(response_cb), tasklist);
+#ifdef HAVE_WAYLAND
+	if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
+		gtk_widget_set_sensitive(tasklist->window_list_content_box, FALSE);
+		gtk_widget_set_sensitive(tasklist->window_grouping_box, FALSE);
+		gtk_widget_set_sensitive(tasklist->minimized_windows_box, FALSE);
+		gtk_widget_set_sensitive(tasklist->mouse_scroll_check, FALSE);
+	#ifdef HAVE_WINDOW_PREVIEWS
+		gtk_widget_set_sensitive(tasklist->window_thumbnail_box, FALSE);
+	#endif
+	}
+#endif /* HAVE_WAYLAND */
 }
 
 static void display_properties_dialog(GtkAction* action, TasklistData* tasklist)
@@ -980,18 +857,27 @@ static void display_properties_dialog(GtkAction* action, TasklistData* tasklist)
 		gtk_builder_set_translation_domain(builder, GETTEXT_PACKAGE);
 		gtk_builder_add_from_resource (builder, WNCKLET_RESOURCE_PATH "window-list.ui", NULL);
 
-		tasklist->properties_dialog = WID("tasklist_properties_dialog");
+		tasklist->properties_dialog = GET_WIDGET("tasklist_properties_dialog");
 
 		g_object_add_weak_pointer(G_OBJECT(tasklist->properties_dialog), (void**) &tasklist->properties_dialog);
 
 		setup_dialog(builder, tasklist);
 
+	gtk_builder_add_callback_symbols (builder,
+#ifdef HAVE_WINDOW_PREVIEWS
+		                              "on_thumbnail_size_spin_value_changed", G_CALLBACK (on_thumbnail_size_spin_value_changed),
+#endif
+		                              "on_never_group_radio_toggled", G_CALLBACK (on_never_group_radio_toggled),
+		                              "on_auto_group_radio_toggled", G_CALLBACK (on_auto_group_radio_toggled),
+		                              "on_always_group_radio_toggled", G_CALLBACK (on_always_group_radio_toggled),
+		                              "on_tasklist_properties_dialog_response", G_CALLBACK (on_tasklist_properties_dialog_response),
+		                              NULL);
+
+		gtk_builder_connect_signals(builder, tasklist);
+
 		g_object_unref(builder);
 	}
 
-	gtk_window_set_icon_name(GTK_WINDOW(tasklist->properties_dialog), WINDOW_LIST_ICON);
-
-	gtk_window_set_resizable(GTK_WINDOW(tasklist->properties_dialog), FALSE);
 	gtk_window_set_screen(GTK_WINDOW(tasklist->properties_dialog), gtk_widget_get_screen(tasklist->applet));
 	gtk_window_present(GTK_WINDOW(tasklist->properties_dialog));
 }
